from __future__ import absolute_import
from .base import BaseModule

import time

class PageTimerModule(BaseModule):
    key = 'overall'

    def __init__(self):
        super(PageTimerModule, self).__init__()
        self._start_time = time.time()

    def get_metrics(self):
        render_time = int((time.time() - self._start_time) * 1000)
        return { 'time' : render_time }


def init():
    return PageTimerModule



from __future__ import absolute_import

try:
    from celery.task import Task as TaskTask
except ImportError:
    TaskTask = None

from .base import BaseModule, RequestTrace
from .stacktracer import trace_method

ENTRY_TYPE = 'CELERY'

class CeleryModule(BaseModule):
    key = 'celery'

    def get_metrics(self):
        return RequestTrace.instance().stacktracer.get_node_metrics(ENTRY_TYPE)

    def get_details(self):
        nodes = RequestTrace.instance().stacktracer.get_nodes(ENTRY_TYPE)
        return [{'type': node.extra['type'], 'args': node.extra['args'], 'kwargs': node.extra['kwargs'], 'time': node.duration} for node in nodes]


def init():
    if TaskTask is None:
        return False

    @trace_method(TaskTask)
    def apply_async(self, args=None, kwargs=None, *_args, **_kwargs):
        return (ENTRY_TYPE, 'Celery: %s' % (self.__name__,), {'type': self.__name__, 'args': args, 'kwargs': kwargs})

    return CeleryModule



from django.utils.importlib import import_module
from django.conf import settings

DETAILS_PREFIX='speedbar:details:'
TRACE_PREFIX='speedbar:trace:'

SPEEDBAR_MODULES = [
    'speedbar.modules.stacktracer', # Most other modules depend on this one
    'speedbar.modules.pagetimer',
    'speedbar.modules.hostinformation',
    'speedbar.modules.sql',
    'speedbar.modules.celeryjobs',
    'speedbar.modules.requeststages',
    'speedbar.modules.templates',
    'speedbar.modules.redis',
    'speedbar.modules.memcache',
    'speedbar.modules.haystack',
    'speedbar.modules.cassandra',
]

# A module comprises of two parts, both of which are optional. It may have an init() function which is called once
# on server startup, and it may have a class called Module which is instantiated once per request.

loaded_modules = []

modules_initialised = False
def init_modules():
    """
    Run the init function for all modules which have one
    """
    global modules_initialised
    if modules_initialised:
        return
    modules_initialised = True

    for module_name in getattr(settings, 'SPEEDBAR_MODULES', SPEEDBAR_MODULES):
        python_module = import_module(module_name)
        speedbar_module = python_module.init()
        if speedbar_module:
            loaded_modules.append(speedbar_module)



from django.conf.urls import *

urlpatterns = patterns('speedbar.views',
    url(r'^panel/(?P<trace_id>[a-zA-Z0-9_-]+)/$', 'panel', name='speedbar_panel'),
    url(r'^trace/(?P<trace_id>[a-zA-Z0-9_-]+)/$', 'trace', name='speedbar_trace'),
    url(r'^details-for-this-request/$', 'noop', name='speedbar_details_for_this_request'),
)



from django.core.cache import cache
from django.dispatch import Signal
from speedbar.utils import DETAILS_PREFIX, TRACE_PREFIX, loaded_modules
from speedbar.modules.base import RequestTrace

DETAILS_CACHE_TIME = 60 * 30  # 30 minutes


request_trace_complete = Signal(providing_args=['metrics', 'request', 'response'])


def setup_request_tracing(sender, **kwargs):
    RequestTrace(module() for module in loaded_modules)
    RequestTrace.instance().stacktracer.push_stack('HTTP', '')


def store_request_trace(sender, **kwargs):
    request_trace = RequestTrace.instance()
    if not request_trace:
        return

    request_trace.stacktracer.pop_stack()

    # Calculate values before doing any cache writes, so the cache writes don't affect the results
    if request_trace.persist_details:
        details_tuples = tuple(
            (key, module.get_details()) for key, module in request_trace.modules.items()
        )
        all_details = dict(details for details in details_tuples if details[1] is not None)
    if request_trace.persist_log:
        speedtracer_log = request_trace.stacktracer.speedtracer_log()
    metrics = dict((key, module.get_metrics()) for key, module in request_trace.modules.items())

    if request_trace.persist_details:
        cache.set(DETAILS_PREFIX + request_trace.id, all_details, DETAILS_CACHE_TIME)
    if request_trace.persist_log:
        cache.set(TRACE_PREFIX + request_trace.id, speedtracer_log, DETAILS_CACHE_TIME)

    request_trace_complete.send(
        sender,
        metrics=metrics,
        request=getattr(request_trace, 'request', None),
        response=getattr(request_trace, 'response', None)
    )



from __future__ import absolute_import

try:
    from cassandra.cluster import Session
except ImportError:
    Session = None

from .base import BaseModule, RequestTrace
from .stacktracer import trace_method


class CassandraModule(BaseModule):
    key = 'cassandra'

    def get_metrics(self):
        return RequestTrace.instance().stacktracer.get_node_metrics('CASSANDRA')

    def get_details(self):
        nodes = RequestTrace.instance().stacktracer.get_nodes('CASSANDRA')
        return [{'cql': node.label, 'time': node.duration} for node in nodes]


def init():
    if Session is None:
        return False

    # The linter thinks the methods we monkeypatch are not used
    # pylint: disable=W0612
    @trace_method(Session)
    def execute(self, query, parameters=None, *args, **kwargs):
        return ('CASSANDRA', query, {})

    return CassandraModule



from uuid import uuid4
import threading


class ThreadLocalSingleton(object):
    def __init__(self):
        if not hasattr(self.__class__, '_thread_lookup'):
            self.__class__._thread_lookup = threading.local()
        self.__class__._thread_lookup.instance = self

    def release(self):
        if getattr(self.__class_._thread_lookup, 'instance', None) is self:
            self.__class_._thread_lookup.instance = None

    @classmethod
    def instance(cls):
        if hasattr(cls, '_thread_lookup'):
            return getattr(cls._thread_lookup, 'instance', None)


class RequestTrace(ThreadLocalSingleton):
    """
    This is a container which keeps track of all module instances for a single request. For convenience they are made
    available as attributes based on their keyname
    """
    def __init__(self, modules=[]):
        super(RequestTrace, self).__init__()
        self.id = str(uuid4())
        self.modules = dict((m.key, m) for m in modules)
        self.__dict__.update(self.modules)
        self.persist_details = False
        self.persist_log = False

class BaseModule(object):
    def get_metrics(self):
        """
        Get a dictionary of summary metrics for the module
        """
        return dict()

    def get_details(self):
        """
        Get a detailed breakdown of all information collected by the module if available
        """
        return None



from __future__ import absolute_import

from collections import defaultdict

from .base import BaseModule, RequestTrace
from .monkey_patching import monkeypatch_method, CallableProxy

import time

class StackEntry(object):
    def __init__(self, id_generator, entry_map, entry_type, label, extra=None):
        self.id_generator = id_generator
        self.entry_map = entry_map
        self.entry_id = id_generator()
        self.entry_type = entry_type
        self.label = label
        self.extra = extra
        self.start = time.time()
        self.children = []
        self.entry_map[entry_type].append(self)

    def mark_end(self):
        self.end = time.time()

    def add_child(self, entry_type, label, extra=None):
        child = StackEntry(self.id_generator, self.entry_map, entry_type, label, extra)
        self.children.append(child)
        return child

    @property
    def duration(self):
        if self.end:
            return self.end - self.start
        return 0

    def to_dict(self):
        return {
            'id': str(self.entry_id),
            'range': {
                'start': round(self.start * 1000, 1),
                'end': round(self.end * 1000, 1),
                'duration': round(self.duration * 1000, 1),
            },
            'operation' : {
                'type': self.entry_type,
                'label': self.label,
            },
            'children': [child.to_dict() for child in self.children],
        }


class StackTracer(BaseModule):
    """
    This class maintains a call tree, with a pointer to the current stack
    entry so that new frames can be added at any time without further context
    by the various monkey patching functions.

    It can provide all entries corresponding to operations of particular types, and
    also build a valid HAR file out of the entire call tree for use with SpeedTracer
    """
    key = 'stacktracer'

    def __init__(self):
        super(StackTracer, self).__init__()
        self.root = None
        self.stack = []
        self.stack_id = 0
        self.entry_map = defaultdict(list)

    def push_stack(self, entry_type, label, extra=None):
        if len(self.stack):
            entry = self.stack[-1].add_child(entry_type, label, extra)
        else:
            entry = self.root = StackEntry(self._get_next_id, self.entry_map, entry_type, label, extra)
        self.stack.append(entry)
        return entry

    def pop_stack(self):
        self.stack[-1].mark_end()
        self.stack.pop()

    def get_metrics(self):
        return {}

    def get_node_metrics(self, node_type):
        nodes = self.get_nodes(node_type)
        return {
            'time': int(sum(x.duration for x in nodes)*1000),
            'count': len(nodes),
        }

    def get_nodes(self, node_type):
        return self.entry_map[node_type]

    def speedtracer_log(self):
        entries_as_dict = self.root.to_dict()
        return {
            'trace': {
                'id': str(self.stack_id),
                'application': 'Speedbar',
                'date': time.time(),
                'range': entries_as_dict['range'],
                'frameStack': entries_as_dict,
            }
        }

    def _get_next_id(self):
        self.stack_id += 1
        return self.stack_id


# The linter thinks the methods we monkeypatch are not used
# pylint: disable=W0612
def trace_method(cls, method_name=None):
    def decorator(info_func):
        method_to_patch = method_name or info_func.__name__
        @monkeypatch_method(cls, method_to_patch)
        def tracing_method(original, self, *args, **kwargs):
            request_trace = RequestTrace.instance()
            if request_trace:
                entry_type, label, extra = info_func(self, *args, **kwargs)
                request_trace.stacktracer.push_stack(entry_type, label, extra=extra)
            try:
                return original(*args, **kwargs)
            finally:
                if request_trace:
                    request_trace.stacktracer.pop_stack()
        return tracing_method
    return decorator


def trace_function(func, info):
    try:
        def tracing_function(original, *args, **kwargs):
            request_trace = RequestTrace.instance()
            if request_trace:
                if callable(info):
                    entry_type, label, extra = info(*args, **kwargs)
                else:
                    entry_type, label, extra = info
                request_trace.stacktracer.push_stack(entry_type, label, extra)
            try:
                return original(*args, **kwargs)
            finally:
                if request_trace:
                    request_trace.stacktracer.pop_stack()
        return CallableProxy(func, tracing_function)
    except Exception:
        # If we can't wrap for any reason, just return the original
        return func


def init():
    return StackTracer



from __future__ import absolute_import

from django import template
register = template.Library()

@register.simple_tag
def metric(module, metric):
    """
    Display a placeholder that the middleware converts to a particular summary metric
    """
    return '<span data-module="%s" data-metric="%s"></span>' % (module, metric)



from __future__ import absolute_import

from django.core import urlresolvers
from django.core.handlers.base import BaseHandler
from django.core.handlers.wsgi import WSGIHandler

from .base import RequestTrace
from .monkey_patching import monkeypatch_method
from .stacktracer import trace_function

import traceback


def patch_function_list(functions, action_type, format_string):
    for i, func in enumerate(functions):
        if hasattr(func, 'im_class'):
            middleware_name = func.im_class.__name__
        else:
            middleware_name = func.__name__
        info = (action_type, format_string % (middleware_name,), {})
        functions[i] = trace_function(func, info)


def wrap_middleware_with_tracers(request_handler):
    patch_function_list(request_handler._request_middleware, 'MIDDLEWARE_REQUEST', 'Middleware: %s (request)')
    patch_function_list(request_handler._view_middleware, 'MIDDLEWARE_VIEW', 'Middleware: %s (view)')
    patch_function_list(request_handler._template_response_middleware, 'MIDDLEWARE_TEMPLATE_RESPONSE', 'Middleware: %s (template response)')
    patch_function_list(request_handler._response_middleware, 'MIDDLEWARE_RESPONSE', 'Middleware: %s (response)')
    patch_function_list(request_handler._exception_middleware, 'MIDDLEWARE_EXCEPTION', 'Middleware: %s (exeption)')


# The linter thinks the methods we monkeypatch are not used
# pylint: disable=W0612
middleware_patched = False
def intercept_middleware():
    @monkeypatch_method(WSGIHandler)
    def __call__(original, self, *args, **kwargs):
        # The middleware cache may have been built before we have a chance to monkey patch
        # it, so do so here
        global middleware_patched
        if not middleware_patched and self._request_middleware is not None:
            self.initLock.acquire()
            try:
                if not middleware_patched:
                    wrap_middleware_with_tracers(self)
                    middleware_patched = True
            finally:
                self.initLock.release()
        return original(*args, **kwargs)

    @monkeypatch_method(BaseHandler)
    def load_middleware(original, self, *args, **kwargs):
        global middleware_patched
        original(*args, **kwargs)
        wrap_middleware_with_tracers(self)
        middleware_patched = True


def intercept_resolver_and_view():
    # The only way we can really wrap the view method is by replacing the implementation
    # of RegexURLResolver.resolve. It would be nice if django had more configurability here, but it does not.
    # However, we only want to replace it when invoked directly from the request handling stack, so we
    # inspect the callstack in __new__ and return either a normal object, or an instance of our proxying
    # class.
    real_resolver_cls = urlresolvers.RegexURLResolver
    class ProxyRegexURLResolverMetaClass(urlresolvers.RegexURLResolver.__class__):
        def __instancecheck__(self, instance):
            # Some places in django do a type check against RegexURLResolver and behave differently based on the result, so we have to
            # make sure the replacement class we plug in accepts instances of both the default and replaced types.
            return isinstance(instance, real_resolver_cls) or super(ProxyRegexURLResolverMetaClass, self).__instancecheck__(instance)
    class ProxyRegexURLResolver(object):
        __metaclass__ = ProxyRegexURLResolverMetaClass
        def __new__(cls, *args, **kwargs):
            real_object = real_resolver_cls(*args, **kwargs)
            stack = traceback.extract_stack()
            if stack[-2][2] == 'get_response':
                obj = super(ProxyRegexURLResolver, cls).__new__(cls)
                obj.other = real_object
                return obj
            else:
                return real_object
        def __getattr__(self, attr):
            return getattr(self.other, attr)
        def resolve(self, path):
            request_trace = RequestTrace.instance()
            if request_trace:
                request_trace.stacktracer.push_stack('RESOLV', 'Resolving: ' + path)
            try:
                callbacks = self.other.resolve(path)
            finally:
                if request_trace:
                    request_trace.stacktracer.pop_stack()
            # Replace the callback function with a traced copy so we can time how long the view takes.
            callbacks.func = trace_function(callbacks.func, ('VIEW', 'View: ' + callbacks.view_name, {}))
            return callbacks
    urlresolvers.RegexURLResolver = ProxyRegexURLResolver


def init():
    intercept_middleware()
    intercept_resolver_and_view()



import os
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as file_handle:
    README = file_handle.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-speedbar',
    version='0.1.3',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='Provides a break down of page loading time',
    long_description=README,
    url='http://github.com/theospears/django-speedbar',
    author='Theo Spears',
    author_email='theo@mixcloud.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'ProxyTypes>=0.9',
        'Django >=1.5, <1.8a0',
    ],
)



from __future__ import absolute_import

from django.db.backends import BaseDatabaseWrapper
from django.db.backends.util import CursorWrapper
from .base import BaseModule, RequestTrace
from .monkey_patching import monkeypatch_method


class SqlModule(BaseModule):
    key = 'sql'

    def __init__(self):
        super(SqlModule, self).__init__()
        self.queries = []

    def get_metrics(self):
        return RequestTrace.instance().stacktracer.get_node_metrics('SQL')

    def get_details(self):
        sql_nodes = RequestTrace.instance().stacktracer.get_nodes('SQL')
        return [{'sql': node.label, 'time': int(node.duration*1000)} for node in sql_nodes]


class _DetailedTracingCursorWrapper(CursorWrapper):
    def execute(self, sql, params=()):
        request_trace = RequestTrace.instance()
        if request_trace:
            stack_entry = request_trace.stacktracer.push_stack('SQL', sql)
        try:
            return self.cursor.execute(sql, params)
        finally:
            if request_trace:
                request_trace.stacktracer.pop_stack()
                sql = self.db.ops.last_executed_query(self.cursor, sql, params)
                stack_entry.label = sql

    def executemany(self, sql, param_list):
        request_trace = RequestTrace.instance()
        if request_trace:
            request_trace.stacktracer.push_stack('SQL', sql)
        try:
            return self.cursor.executemany(sql, param_list)
        finally:
            if request_trace:
                request_trace.stacktracer.pop_stack()


# The linter thinks the methods we monkeypatch are not used
# pylint: disable=W0612
def init():
    @monkeypatch_method(BaseDatabaseWrapper)
    def cursor(original, self, *args, **kwargs):
        result = original(*args, **kwargs)
        return _DetailedTracingCursorWrapper(result, self)

    return SqlModule



# This package has missing __init__.py files, so pylint can't see it
# pylint: disable=F0401
from peak.util.proxies import ObjectWrapper


# The linter is dumb
# pylint: disable=E1001,E1002
class CallableProxy(ObjectWrapper):
    __slots__ = ('_eop_wrapper_')

    def __init__(self, wrapped, wrapper):
        super(CallableProxy, self).__init__(wrapped)
        self._eop_wrapper_ = wrapper

    def __call__(self, *args, **kwargs):
        return self._eop_wrapper_(self.__subject__, *args, **kwargs)


class BoundMethodProxy(ObjectWrapper):
    __slots__ = ('_eop_wrapper_', '_eop_instance_')

    def __init__(self, wrapped, instance, wrapper):
        super(BoundMethodProxy, self).__init__(wrapped)
        self._eop_instance_ = instance
        self._eop_wrapper_ = wrapper

    def __call__(self, *args, **kwargs):
        return self._eop_wrapper_(self.__subject__, self._eop_instance_, *args, **kwargs)


class UnboundMethodProxy(CallableProxy):
    __slots__ = ('_eop_wrapper_')

    def __get__(self, instance, owner):
        return BoundMethodProxy(self.__subject__.__get__(instance, owner), instance or owner, self._eop_wrapper_)

    def __getattribute__(self, attr, oga=object.__getattribute__):
        """We need to return our own version of __get__ or we may end up never being called if
        the member we are wrapping is wrapped again by someone else"""
        if attr == '__get__':
            return oga(self, attr)
        return super(UnboundMethodProxy, self).__getattribute__(attr)


def monkeypatch_method(cls, method_name=None):
    def decorator(func):
        method_to_patch = method_name or func.__name__
        original = cls.__dict__[method_to_patch]
        replacement = UnboundMethodProxy(original, func)
        type.__setattr__(cls, method_to_patch, replacement)  # Avoid any overrides
        return func
    return decorator



from __future__ import absolute_import

try:
    import haystack
    from haystack.exceptions import MissingDependency
except ImportError:
    haystack = None
    MissingDependency = None

from .base import BaseModule, RequestTrace
from .stacktracer import trace_method


ENTRY_TYPE='haystack'

class HaystackModule(BaseModule):
    key = 'haystack'

    def get_metrics(self):
        return RequestTrace.instance().stacktracer.get_node_metrics(ENTRY_TYPE)

    def get_details(self):
        redis_nodes = RequestTrace.instance().stacktracer.get_nodes(ENTRY_TYPE)
        return [{'query_string': node.extra['query_string'], 'kwargs': node.extra['kwargs'], 'time': node.duration} for node in redis_nodes]

def init():
    if haystack is None:
        return False

    def search(self, query_string, *args, **kwargs):
        models = kwargs.get('models', None)
        if models:
            description = '[%s] %s' % (", ".join(m.__name__ for m in models), query_string)
        else:
            description = '[no models specified] %s' % (query_string,)

        return (ENTRY_TYPE, 'Haystack: %s' % (description,), {'query_string' : query_string, 'kwargs': kwargs})

    try:
        from haystack.backends.elasticsearch_backend import ElasticsearchSearchBackend
    except MissingDependency:
        pass
    else:
        trace_method(ElasticsearchSearchBackend)(search)

    try:
        from haystack.backends.simple_backend import SimpleSearchBackend
    except MissingDependency:
        pass
    else:
        trace_method(SimpleSearchBackend)(search)

    try:
        from haystack.backends.solr_backend import SolrSearchBackend
    except MissingDependency:
        pass
    else:
        trace_method(SolrSearchBackend)(search)

    try:
        from haystack.backends.whoosh_backend import WhooshSearchBackend
    except MissingDependency:
        pass
    else:
        trace_method(WhooshSearchBackend)(search)

    return HaystackModule
